#!/bin/bash
#

TOOL_DIR_BIN="${TOOL_DIR}/bin"
TOOL_DATA_DIR="${TOOL_DIR}/data"
WORKING_DIR="/tmp/sign_tool"

# Files
#######

# IPK
IPK_ITB_FILENAME="linux.itb"
IPK_ITBS_FILENAME="${IPK_ITB_FILENAME}-signed"
IPK_CONTROL_GZ_FILENAME="control.tar.gz"
IPK_DATA_GZ_FILENAME="data.tar.gz"
IPK_DEBIANBIN_FILENAME="debian-binary"
IPK_FILES="${IPK_CONTROL_GZ_FILENAME} ${IPK_DATA_GZ_FILENAME} ${IPK_DEBIANBIN_FILENAME}"
IPK_CONTROL_FILENAME="control"
IPK_POSTINST_FILENAME="postinst"

# ITB
ITB_KERNEL_FILENAME="zImage"
ITB_DTB_FILENAME="devicetree.dtb"
ITB_INITRDGZ_FILENAME="initrd.cpio.gz"
ITB_ITS_FILENAME="fitimage.its"

# TZsw
TZSW_TZSW_FILENAME="TZsw.img"
TZSW_TZSWS_FILENAME="${TZSW_TZSW_FILENAME}-signed"

# U-Boot
UBOOT_UBOOT_FILENAME="u-boot.img"
UBOOT_UBOOTS_FILENAME="${UBOOT_UBOOT_FILENAME}-signed"

# SPL
SPL_SPL_FILENAME="SPL"
SPL_SPLS_FILENAME="${SPL_SPL_FILENAME}-signed"

# ROOTFS
ROOTFS_SQFS_FILENAME="keros.sqfs"
ROOTFS_SQFSS_FILENAME="${ROOTFS_SQFS_FILENAME}.sig"

# Scripts/Bin
#######
SCRIPTS_SIGNBIN="${TOOL_DIR_BIN}/signbin.sh"
SCRIPTS_SIGNIMX="${TOOL_DIR_BIN}/signimx.sh"
SCRIPTS_SIGNGPG="${TOOL_DIR_BIN}/signgpg.sh"
BIN_MKIMAGE="${TOOL_DIR_BIN}/mkimage"
BIN_DUMPIMAGE="${TOOL_DIR_BIN}/dumpimage"



### Generic functions
##################

prepare_working_dir()
{
	echo "Preparing working dir..."
	
	# Clean a little the working dir
	[ -d ${WORKING_DIR} ] || mkdir -p ${WORKING_DIR}
	rm -fR ${WORKING_DIR}/*
	
	echo "Working dir prepared"
}

### ROOTFS functions
##################

#
# Sign a file by GPG
#
# $1: The file to sign
# $2: The file to save the signature 
# $3: GPG key to use
#
gpg_sign()
{
	IN_FILE="$1"
	IN_SIGNATURE_FILE="$2"
	IN_GPG_KEY="$3"
	
	echo "Signing file \"${IN_FILE}\" by GPG..."
	
	${SCRIPTS_SIGNGPG} ${IN_FILE} ${IN_GPG_KEY}
	if [ $? -ne 0 ]
	then
		echo "Error when signing file by GPG"
		exit 1
	fi
	
	[[ "${IN_FILE}.sig" -ef "${IN_SIGNATURE_FILE}" ]] || mv "${IN_FILE}.sig" "${IN_SIGNATURE_FILE}"
	echo "GPG signature file: ${IN_SIGNATURE_FILE}"
	
}

### ITB functions
##################

#
# Extract a keros ITB file
#
# $1: The ITB file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple itb in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
itb_extract()
{
	IN_ITB_FILE="$1"
	IN_WORKSPACE="$2"
	
	echo "Extracting Linux ITB \"${IN_ITB_FILE}\" in workspace ${IN_WORKSPACE}..."
	
	ITB_DIR="${WORKING_DIR}/${IN_WORKSPACE}/itb"
	
	# ITB directories
	mkdir -p ${ITB_DIR}
	ITB_KERNEL_FILE="${ITB_DIR}/${ITB_KERNEL_FILENAME}"
	ITB_DTB_FILE="${ITB_DIR}/${ITB_DTB_FILENAME}"
	ITB_INITRDGZ_FILE="${ITB_DIR}/${ITB_INITRDGZ_FILENAME}"
	
	# Extracting Kernel
	echo "Extracting Kernel..."
	${BIN_DUMPIMAGE} -i ${IN_ITB_FILE} -T flat_dt -p 0 ${ITB_KERNEL_FILE}
	if [ $? -ne 0 ]
	then
		echo "Error when extracting kernel from ITB"
		exit 1
	fi
	echo "Kernel extracted"
	
	# Extracting DTB
	echo "Extracting DTB..."
	${BIN_DUMPIMAGE} -i ${IN_ITB_FILE} -T flat_dt -p 1 ${ITB_DTB_FILE}
	if [ $? -ne 0 ]
	then
		echo "Error when extracting DTB from ITB"
		exit 1
	fi
	echo "DTB extracted"
	
	# Extracting INITRD
	echo "Extracting INITRD..."
	${BIN_DUMPIMAGE} -i ${IN_ITB_FILE} -T flat_dt -p 2 ${ITB_INITRDGZ_FILE}
	if [ $? -ne 0 ]
	then
		echo "Error when extracting INITRD from ITB"
		exit 1
	fi
	echo "INITRD extracted"
	
	echo "Linux ITB extracted"
}

#
# Build a Linux ITB from a dir with 3 elements:
# - DTB: devicetree.dtb
# - INITRD: initrd.cpio.gz
# - KERNEL: zImage
#
# $1: The workspace to use as input dir.
# $2: The final initrd file with cpio.gz extension
#
itb_build()
{
	IN_WORKSPACE="$1"
	IN_ITB_FILE="$2"

	echo "Building Linux ITB from workspace ${IN_WORKSPACE}..."
	
	itb_dir="${WORKING_DIR}/${IN_WORKSPACE}/itb"
	
	cd ${itb_dir}
	rm -f ${IN_ITB_FILE}
	cp -f "${TOOL_DATA_DIR}/${ITB_ITS_FILENAME}" ${itb_dir}
	${BIN_MKIMAGE} -A arm -f ${ITB_ITS_FILENAME} ${IN_ITB_FILE}
	if [ $? -ne 0 ]
	then
		echo "Error when creating Linux ITB"
		exit 1
	fi
	
	echo "Linux ITB created: ${IN_ITB_FILE}"
}

#
# Sign a Linux ITB
#
# $1: The Linux ITB file to sign
# $2: The file to save the signed ITB
#
itb_sign()
{
	IN_ITB_FILE="$1"
	IN_ITBS_FILE="$2"
	echo "Signing Linux ITB file \"${IN_ITB_FILE}\"..."
	
	${SCRIPTS_SIGNBIN} ${IN_ITB_FILE} "${FITIMAGE_LOADADDR}"
	if [ $? -ne 0 ]
	then
		echo "Error when signing Linux ITB"
		exit 1
	fi
	
	[[ "${IN_ITB_FILE}-signed" -ef "${IN_ITBS_FILE}" ]] || mv "${IN_ITB_FILE}-signed" "${IN_ITBS_FILE}"
	echo "Linux ITB signed: ${IN_ITBS_FILE}"
}

### TZsw functions
##################

#
# Sign a TZSW
#
# $1: The TZSW file to sign
# $2: The file to save the signed TZSW
#
tzsw_sign()
{
	IN_TZSW_FILE="$1"
	IN_TZSWS_FILE="$2"
	
	echo "Signing TZsw file \"${IN_TZSW_FILE}\"..."
		
	${SCRIPTS_SIGNBIN} ${IN_TZSW_FILE} "${TZSW_LOADADDR} - 0x40"
	if [ $? -ne 0 ]
	then
		echo "Error when signing TZsw"
		exit 1
	fi
	
	[[ "${IN_TZSW_FILE}-signed" -ef "${IN_TZSWS_FILE}" ]] || mv "${IN_TZSW_FILE}-signed" "${IN_TZSWS_FILE}"
	echo "TZsw signed: ${IN_TZSWS_FILE}"
}

### U-Boot functions
##################

#
# Sign a U-Boot
#
# $1: The U-Boot file to sign
# $2: The file to save the signed U-Boot
#
uboot_sign()
{
	IN_UBOOT_FILE="$1"
	IN_UBOOTS_FILE="$2"
	echo "Signing U-Boot file \"${IN_UBOOT_FILE}\"..."
		
	${SCRIPTS_SIGNBIN} ${IN_UBOOT_FILE} "${UBOOT_LOADADDR} - 0x40"
	if [ $? -ne 0 ]
	then
		echo "Error when signing U-Boot"
		exit 1
	fi
	
	[[ "${IN_UBOOT_FILE}-signed" -ef "${IN_UBOOTS_FILE}" ]] || mv "${IN_UBOOT_FILE}-signed" "${IN_UBOOTS_FILE}"
	echo "U-Boot signed: ${IN_UBOOTS_FILE}"
}

### INITRD functions
##################

#
# Extract a cpio.gz INITRD
#
# $1: The INITRD in cpio.gz format file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
initrd_extract()
{
	IN_INITRD_FILE="$1"
	IN_WORKSPACE="$2"
	
	echo "Extracting INITRD \"${IN_INITRD_FILE}\" in workspace ${IN_WORKSPACE}..."

	# INITRD dir
	initrd_dir="${WORKING_DIR}/${IN_WORKSPACE}/initrd"
	mkdir -p ${initrd_dir}
	
	cd ${initrd_dir}
	gzip -dc ${IN_INITRD_FILE} | cpio -idm
	if [ $? -ne 0 ]
	then
		echo "Error when extracting INITRD"
		exit 1
	fi
	echo "INITRD extracted"
	
}

#
# Deploy GPG keys inside an initrd
#
# $1: The tar.gz file with gpg keys to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
initrd_install_gpg_keys()
{
	IN_GPG_KEYS_FILE="$1"
	IN_WORKSPACE="$2"
	
	echo "Installing GPG keys \"${IN_GPG_KEYS_FILE}\" in INITRD..."
	
	# INITRD dir
	initrd_dir="${WORKING_DIR}/${IN_WORKSPACE}/initrd"
	
	cd ${initrd_dir}
	mkdir -p -m 700 root
	tar -C root -xpf ${IN_GPG_KEYS_FILE}
	
	echo "GPG keys installed"
}

#
# Build an INITRD from a dir
#
# $1: The workspace to use as input dir.
# $1: The final initrd file with cpio.gz extension
#
initrd_build()
{
	IN_WORKSPACE="$1"
	IN_INITRD_FILE="$2"

	echo "Building INITRD from workspace ${IN_WORKSPACE}..."
	
	initrd_dir="${WORKING_DIR}/${IN_WORKSPACE}/initrd"
	
	cd ${initrd_dir}
	rm -f ${IN_INITRD_FILE} 
	find . | cpio --create --format='newc' | gzip > ${IN_INITRD_FILE}
	
	echo "INITRD created: ${IN_INITRD_FILE}"
	
}

### IPK functions
##################

#
# Extract an IPK package
#
# $1: The IPK file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
ipk_extract()
{
	IN_IPK="$1"
	IN_WORKSPACE="$2"
	
	echo "Extracting IPK file \"${IN_IPK}\" in workspace ${IN_WORKSPACE}..."
	
	ipk_file="`basename ${IN_IPK}`"
	ipk_dir="${WORKING_DIR}/${IN_WORKSPACE}/ipk"
	ipk_data_dir="${ipk_dir}/data"
	ipk_control_dir="${ipk_dir}/control"
	
	# IPK directory
	mkdir -p ${ipk_data_dir} ${ipk_control_dir}
	cp -f ${IN_IPK} ${ipk_dir}/
	cd ${ipk_dir}
	
	# Uncompress IPK
	ar x ${ipk_file}

	# Uncompress data part
	tar -xzf ${IPK_DATA_GZ_FILENAME} -C ${ipk_data_dir}
	# Uncompress control part
	tar -xzf ${IPK_CONTROL_GZ_FILENAME} -C ${ipk_control_dir} 
	
	rm -f ${IPK_DATA_GZ_FILENAME} ${IPK_CONTROL_GZ_FILENAME}
	echo "IPK file \"${IN_IPK}\" extracted"
}

#
# Adapt control file of IPK for a signed package
#
# $1: The workspace to use as input to adapt control file
# $2: The version of the package to force
#
ipk_adapt_control()
{
	IN_WORKSPACE="$1"
	IN_IPK_VERSION="$2"
	
	ipk_dir="${WORKING_DIR}/${IN_WORKSPACE}/ipk"
	control_file="${ipk_dir}/control/${IPK_CONTROL_FILENAME}"
	postinst_file="${ipk_dir}/control/${IPK_POSTINST_FILENAME}"

	arch=$(grep Architecture ${control_file} | awk '{print $2}')

	# Adapt control file
	sed -i "s/Version: .*$/Version: ${IN_IPK_VERSION}/" ${control_file}
	sed -i '/^Description:/,+2d' ${control_file}
	cat >> ${control_file} << EOF
Description: Kerlink OS
   Kerlink OS tunned for ${arch} target
EOF
	sed -i 's/keros-neutral/keros/g' ${control_file}
	
	# Adapt postinst script
	sed -i 's/SPL/SPL-signed/g' ${postinst_file}
	sed -i 's/u-boot.img/u-boot.img-signed/g' ${postinst_file}
	sed -i 's/TZsw.img/TZsw.img-signed/g' ${postinst_file}

}

#
# Build a IPK package from a list of elements:
# - debian-binary
# - data.tar.gz
# - control.tar.gz
#
# $1: The workspace to use as input to generate the IPK
# $2: The result IPK file
ipk_build()
{
	IN_WORKSPACE="$1"
	IN_IPK_FILE="$2"
	
	echo "Building IPK \"${IN_IPK_FILE}\" from workspace ${IN_WORKSPACE}..."
	
	ipk_dir="${WORKING_DIR}/${IN_WORKSPACE}/ipk"
	ipk_data_dir="${ipk_dir}/data"
	ipk_control_dir="${ipk_dir}/control"
	ipk_data_gz_file="${ipk_dir}/${IPK_DATA_GZ_FILENAME}"
	ipk_control_gz_file="${ipk_dir}/${IPK_CONTROL_GZ_FILENAME}"
	
	cd ${ipk_dir}
	## DATA part
	#############
	
	# Recompress data part
	tar --owner=root --group=root -czf ${ipk_data_gz_file} -C ${ipk_data_dir} . 
	if [ $? -ne 0 ]
	then
		echo "Error when compressing data part"
		exit 1
	fi
	
	## CONTROL part
	#############
	
	# Recompress control part
	tar --owner=root --group=root -czf ${ipk_control_gz_file} -C ${ipk_control_dir} . 
	if [ $? -ne 0 ]
	then
		echo "Error when compressing control part"
		exit 1
	fi
	
	## IPK Creation
	#############
	ar r ${IN_IPK_FILE} ${IPK_FILES}
	if [ $? -ne 0 ]
	then
		echo "Error when creating IPK"
		exit 1
	fi
	
	echo "IPK built: ${IN_IPK_FILE}"
}

### SPL functions
##################

#
# Sign a SPL
#
# $1: The SPL file to sign
# $2: The file to save the signed SPL
#
spl_sign()
{
	IN_SPL_FILE="$1"
	IN_SPLS_FILE="$2"
	
	echo "Signing SPL file \"${IN_SPL_FILE}\"..."
		
	${SCRIPTS_SIGNIMX} ${IN_SPL_FILE} "${SPL_HAB_BLOCKS}"
	if [ $? -ne 0 ]
	then
		echo "Error when signing SPL"
		exit 1
	fi
	
	[[ "${IN_SPL_FILE}-signed" -ef "${IN_SPLS_FILE}" ]] || mv "${IN_SPL_FILE}-signed" "${IN_SPLS_FILE}"
	echo "SPL signed: ${IN_SPLS_FILE}"
}


### Other functions
##################

#
# Load Signature information 
#
# $1: The signinfo file from a neutral ipk
#
load_signinfo()
{
	IN_SIGNINFO_FILE="$1"
	
	# Load Signature information
	if ! [ -e ${IN_SIGNINFO_FILE} ]
	then
		echo "No signature information file found!"
		exit 1
	fi
	. ${IN_SIGNINFO_FILE}
}
