#!/bin/bash -e
#
# Signing tools and update ipks


# Directories
#############
TOOL_DIR=`dirname $0`
TOOL_DIR=`readlink -f ${TOOL_DIR}`
LIB_FILE="tools.sh"

#include signing lib
. ${TOOL_DIR}/${LIB_FILE}

#WORKSPACES
LIVEBURNER_WORKSPACE="liveburner"
KEROS_WORKSPACE="keros"
BURNER_WORKSPACE="burner"

GPG_USER="customer"
KEROS_IPK="keros.ipk"

#Signing information
BLK_SIZE=$((0x100000))
LOADADDR=0x87800000
SIGNATURE_PACKAGE_FILES="linux-burner.itb u-boot.bin imximage.cfg"

### Other functions
##################


#
# Fully sign an itb file by extracting it's content
# and signing every needed file in it. The itb file is then
# rebuild and sign
#
# $1: The itb file to sign
# $2: The workspace to use to extract the content.
#
sign_itb_file()
{
	local in_itb_file="$(realpath $1)"
	local in_workspace="$2"
	local itbfile_dir="dirname ${in_itb_file}"
	local keros_ipk_dir="${WORKING_DIR}/${in_workspace}/initrd/burner"

	#extract itb and initrd
	itb_extract "${in_itb_file}" "${in_workspace}"
	initrd_extract "${WORKING_DIR}/${in_workspace}/itb/${ITB_INITRDGZ_FILENAME}" "${in_workspace}"

	# Check for gpg key (already sign file)
	if [ -d ${WORKING_DIR}/${in_workspace}/initrd/root ]
	then
		echo "Error : File ${FILENAME} is already sign"
		#Cleanup
		rm -rf ${WORKING_DIR}
		exit 1
	fi

	# Sign the keros ipk for the liveburner
	if [ -d "${keros_ipk_dir}" ]
	then
		#Check if keros ipk has already been sign
		if [ ! -f "${IPK_KEROSS_FILE}" ]
		then
			sign_keros_ipk "${keros_ipk_dir}/${KEROS_IPK}" "${WORKING_DIR}"
		fi
		cp -f "${IPK_KEROSS_FILE}" "${keros_ipk_dir}/${KEROS_IPK}"
	fi

	initrd_install_gpg_keys  "${INITRD_GPG_KEYS_FILENAME}" "${in_workspace}"
	initrd_build "${in_workspace}" "${WORKING_DIR}/${in_workspace}/itb/${ITB_INITRDGZ_FILENAME}"

	# Building ITB
	itb_build "${in_workspace}" "${in_itb_file}"

	# Signatures
	itb_sign ${in_itb_file} ${in_itb_file}
}


#
# Fully sign an ipk file by extracting it's content
# and signing every needed file in it. The ipk file is then
# rebuild and sign
#
# $1: The file to be sign
# $2: The directory for the output signed ipk
#
sign_keros_ipk()
{
	# Get original IPK file
	local in_ipk_file="$(readlink -f $1)"
	local in_output_dir="$(readlink -f $2)"
	local ipk_keros_data_dir="${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/keros"
	local ipk_tmp_data_dir="${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/tmp"

	# Extracting the ipk
	ipk_extract "${in_ipk_file}" "${KEROS_WORKSPACE}"

	#Sign the required files
	sign_itb_file "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/keros/${IPK_ITB_FILENAME}" "${KEROS_WORKSPACE}"
	tzsw_sign "${ipk_tmp_data_dir}/${TZSW_TZSW_FILENAME}" "${ipk_tmp_data_dir}/${TZSW_TZSW_FILENAME}"
	uboot_sign "${ipk_tmp_data_dir}/${UBOOT_UBOOT_FILENAME}" "${ipk_tmp_data_dir}/${UBOOT_UBOOT_FILENAME}"
	spl_sign "${ipk_tmp_data_dir}/${SPL_SPL_FILENAME}" "${ipk_tmp_data_dir}/${SPL_SPL_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/tmp/${SPL_SIGNINFO}"
	gpg_sign "${ipk_keros_data_dir}/${ROOTFS_SQFS_FILENAME}" "${ipk_keros_data_dir}/${ROOTFS_SQFS_FILENAME}.sig" "${GPG_USER}"

	#Adapt the ipk control with the signed files
	local control_file="${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/control/${IPK_CONTROL_FILENAME}"
	IPK_VERSION=$(grep Version ${control_file} | awk '{print $2}')
	ipk_adapt_control "${KEROS_WORKSPACE}" "${IPK_VERSION}"

	# IPK content is ready, generate the IPK itself
	IPK_ARCH=$(grep Architecture ${control_file} | awk '{print $2}'| tr _ -)
	IPK_KEROSS_FILE="${in_output_dir}/$(basename $(echo ${in_ipk_file} | sed 's:\(.*keros.*\)-unsigned.ipk:\1-signed.ipk:'))"

	# Extracting the ipk
	ipk_extract "${in_ipk_file}" "${KEROS_WORKSPACE}-signed"

	#Copy the required signed files the required files
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/keros/${IPK_ITB_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/keros/${IPK_ITB_FILENAME}"
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/tmp/${TZSW_TZSW_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/tmp/${TZSW_TZSW_FILENAME}"
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/tmp/${UBOOT_UBOOT_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/tmp/${UBOOT_UBOOT_FILENAME}"
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/tmp/${SPL_SPL_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/tmp/${SPL_SPL_FILENAME}"
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/data/keros/${ROOTFS_SQFS_FILENAME}.sig" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/keros/${ROOTFS_SQFS_FILENAME}.sig"
	cp "${WORKING_DIR}/${KEROS_WORKSPACE}/ipk/control/${IPK_CONTROL_FILENAME}" "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/control/${IPK_CONTROL_FILENAME}"

	#Remove SPL_signinfo from IPK
	rm "${WORKING_DIR}/${KEROS_WORKSPACE}-signed/ipk/data/tmp/${SPL_SIGNINFO}"

	ipk_build "${KEROS_WORKSPACE}-signed" "${IPK_KEROSS_FILE}"

	echo "Signed IPK is built: ${IPK_KEROSS_FILE}"
}


#
# Fully sign a liveburner file by extracting it's containt
# and signing every needed file in it. The liveburner file is then
# rebuild and sign
#
# $1: The liveburner file to be sign
# $2: The output directory for the signed liveburner
#
sign_liveburner_ipk()
{
	# Get original IPK file
	local in_liveburner_file="$(readlink -f $1)"
	local liveburner_dir=$(dirname ${in_liveburner_file})

	# Extracting the ipk
	ipk_extract "${in_liveburner_file}" "${LIVEBURNER_WORKSPACE}"

	#Sign the required files
	sign_itb_file "${WORKING_DIR}/${LIVEBURNER_WORKSPACE}/ipk/data/tmp/${IPK_ITB_FILENAME}" "${LIVEBURNER_WORKSPACE}"

	#Copy signature package file if exist
	for file in ${SIGNATURE_PACKAGE_FILES}
	do
		local file_path="${WORKING_DIR}/${LIVEBURNER_WORKSPACE}/ipk/data/tmp/${file}"
		[ -f "${file_path}" ] && mv "${file_path}" "${WORKING_DIR}"
	done

	#Adapt the ipk control with the signed files
	local control_file="${WORKING_DIR}/${LIVEBURNER_WORKSPACE}/ipk/control/${IPK_CONTROL_FILENAME}"
	local ipk_version=$(grep Version ${control_file} | awk '{print $2}')
	ipk_adapt_control "${LIVEBURNER_WORKSPACE}" "${ipk_version}"

	# IPK content is ready, generate the IPK itself
	local arch=$(grep Architecture ${control_file} | awk '{print $2}')
	local ipk_liveburner_signed_file=$(echo ${in_liveburner_file} | sed 's:.*\(liveburner.*\)-unsigned.ipk:\1-signed.ipk:')
	ipk_build "${LIVEBURNER_WORKSPACE}" "${ipk_liveburner_signed_file}"

	cp "${ipk_liveburner_signed_file}" "${liveburner_dir}/"
	echo "Signed liveburner is built: ${liveburner_dir}/${ipk_liveburner_signed_file}"
}


#
# Generate a sign burner from a build directory
#
# $1: The directory containing required file (usually tmp/deploy/images/klk-lpbs/)
#
generate_burner()
{
	local burner_dir=`readlink -f $1`
	local burner_itb="linux-burner.itb"
	local file_buffer=burner

	#Prepare the workspace
	mkdir -p "${WORKING_DIR}/${BURNER_WORKSPACE}"
	cp "${burner_dir}/${burner_itb}" "${WORKING_DIR}/${BURNER_WORKSPACE}/${IPK_ITB_FILENAME}"

	#Sign the required files
	sign_itb_file "${WORKING_DIR}/${BURNER_WORKSPACE}/${IPK_ITB_FILENAME}" "burner"
	cd ..

	#Process the file sizes and padding
	local binary_size=$(wc -c ${burner_dir}/u-boot.bin | awk '{print $1}')
	local total_size=$(( (${binary_size} / ${BLK_SIZE} + 1) * ${BLK_SIZE} ))
	local pad_size=$((${total_size}-${binary_size}))
	local fitimage_addr=$(($LOADADDR + $total_size))

	#Add padding to the u-boot file
	cp -L "${burner_dir}/u-boot.bin" "${file_buffer}"
	echo "Padding ${file_buffer} with $((${pad_size})) bytes"
	dd if=/dev/zero of="${file_buffer}" oflag=append conv=notrunc bs=1 count=${pad_size} > /dev/null 2>&1

	# Concatenate the itb file to the padded file
	printf "concatenate $FITIMAGE at 0x%X (0x%X)\n" $total_size $fitimage_addr
	cat "${WORKING_DIR}/${BURNER_WORKSPACE}/${IPK_ITB_FILENAME}" >> ${file_buffer}

	# make the .imx file
	${BIN_MKIMAGE} -n "${burner_dir}/imximage.cfg" -T imximage -e ${LOADADDR} -d $file_buffer "${file_buffer}.imx"
	rm -f "${file_buffer}"

	#sign the imx file
	local header_program_hab_block=$(${SCRIPTS_IMX_INFO} --header-program-hab ${file_buffer}.imx)
	local mfg_dcd_hab_block=$(${SCRIPTS_IMX_INFO} --mfg-dcd-hab ${file_buffer}.imx)
	imx_sign "${WORKING_DIR}/${BURNER_WORKSPACE}/${file_buffer}.imx" "${header_program_hab_block}, ${mfg_dcd_hab_block}"

	cp "${WORKING_DIR}/${BURNER_WORKSPACE}/${file_buffer}.imx-signed" "${burner_dir}/${file_buffer}_${IPK_VERSION}_${IPK_ARCH}-signed.imx"

	echo "Signed burner is built: ${burner_dir}/${file_buffer}_${IPK_VERSION}_${IPK_ARCH}-signed.imx"
}

#
# Sign every files generated from the last build.
# The function use the symbolique link created
# during a build to identify the last image build.
#
# $1: The directory containing required file (usually tmp/deploy/images/klk-lpbs/)
#
sign_all()
{
	local file_dir=`readlink -f $1`

	#retrieve machine architecture
	local machine=`head -n 1 ${file_dir}/imximage.cfg | awk -F "/" '{print $4}'`

	#Find all keros ipk (base ipk + test ones) and sort in reverse order to keep base ipk value for liveburner and burner
	local keros_file=`readlink -f ${file_dir}/keros_${machine}.ipk`
	if [ -z "${keros_file##*test*}" ]
	then
		local ipk_base_name=$(echo ${keros_file} | sed 's:.*\(keros.*\)-test.*:\1:')
		local ipk_files="$(find "${file_dir}" -name "${ipk_base_name}*${machine}-unsigned.ipk" | sort -nr)"
	else
		local ipk_files="${keros_file}"
	fi

	#find the liveburner legacy
	local liveburner_file=`readlink ${file_dir}/liveburner_${machine}.ipk`
	local liveburner_file_suffix="${liveburner_file#liveburner}"
	local legacy_file=$(find "${file_dir}" -name "liveburner-legacy${liveburner_file_suffix}")
	for file in ${ipk_files}
	do
		#clear the directory from previous ipk data
		rm -rf "${WORKING_DIR}/${KEROS_WORKSPACE}"
		sign_keros_ipk "${file}" "${file_dir}"
	done

	#sign liveburner
	sign_liveburner_ipk "${file_dir}/${liveburner_file}"

	#clean before signing the legacy liveburner
	rm -rf "${WORKING_DIR}/${LIVEBURNER_WORKSPACE}"
	#sign the legacy liveburner
	sign_liveburner_ipk "${legacy_file}"


	#Sign burner file
	generate_burner "${file_dir}"

}

#
# Sign a single file. Only used for development.
# The function detects the type of file using the filename:
#   - SPL*: A SPL file
#   - TZsw*: A Prove&Core Software
#   - u-boot*.img*: A U-Boot image
#   - *.itb: A SPL file
#   - *.sqfs: A KerOS Squash file
#
# $1: The file to sign (i.e. tmp/deploy/images/klk-lpbs/u-boot.img)
#
sign_file()
{
	local in_file="$1"
	local resolved_filename="$(readlink -f ${in_file})"
	local dirname="$(dirname ${resolved_filename})"
	local filename=`basename "${resolved_filename}"`

	case ${filename} in
		SPL*)
			echo "SPL file detected"
			spl_sign "${resolved_filename}" "${resolved_filename}" "${dirname}/${SPL_SIGNINFO}"
			;;
		TZsw*)
			echo "TZsw file detected"
			tzsw_sign "${resolved_filename}" "${resolved_filename}"
			;;
		u-boot*.img*)
			echo "U-boot file detected"
			uboot_sign "${resolved_filename}" "${resolved_filename}"
			;;
		*.itb*)
			echo "ITB file detected"
			sign_itb_file "${resolved_filename}" "${KEROS_WORKSPACE}"
			;;
		*.sqfs*|*.squashfs)
			echo "KerOS Squash file detected"
			gpg_sign "${resolved_filename}" "${in_file}.sig" "${GPG_USER}"
			;;
		*)
			echo "Unsupported file"
			;;
	esac

}

#
# Generate all signed file from keros signature package file.
# The generated files are : keros.ipk, liveburner.ipk, burner.imx
#
# $1: The keros signature package file
#
sign_package()
{
	local in_file=`readlink -f $1`
	local file_dir=`dirname ${in_file}`
	local file_suffix=$(echo "${in_file}" | sed 's:.*keros_signature_package_\(.*\):\1:')
	local liveburner_filename="liveburner_${file_suffix}-unsigned.ipk"
	local keros_filename="keros_${file_suffix}-signed.ipk"
	local burner_filename="burner_${file_suffix}-signed.imx"

	cp "${in_file}" "${WORKING_DIR}/${liveburner_filename}"
	sign_liveburner_ipk "${WORKING_DIR}/${liveburner_filename}" "${LIVEBURNER_WORKSPACE}"
	generate_burner "${WORKING_DIR}"

	#copy the builded files
	mv "${WORKING_DIR}/${KEROS_IPK}" "${file_dir}/${keros_filename}"
	mv "${WORKING_DIR}/${liveburner_filename%-unsigned.ipk}-signed.ipk" "${file_dir}/${liveburner_filename%-unsigned.ipk}-signed.ipk"
	mv $(find "${WORKING_DIR}" -maxdepth 1 -type f -name "burner*") "${file_dir}/${burner_filename}"
  printf "Signed files have been generated : \nburner : ${file_dir}/${burner_filename}\n"
  printf "liveburner : ${file_dir}/${liveburner_filename%-unsigned.ipk}-signed.ipk\nkeros : ${file_dir}/${keros_filename}\n"
}

usage()
{
	echo -e "Usage: $0 [option] [file]
\t-a/--all [directory] : Sign all necessary file for a release (ipk, liveburners and burner) from a release directory
\t-b/--burner [directory] : Generate a signed burner from release directory
\t-f/--file [file] : Sign a single file (for development only): SPL, u-boot, TZ, ITB, KerOS Squash
\t-h/--help : display this help
\t-k/--keros [ipk file] : Sign a keros ipk file
\t-l/--liveburner [liveburner file] : Sign a liveburner file
\t-p/--package [signature package] : Generate keros.ipk liveburner.ipk and burner from a signature package file"
}


##################
#      MAIN
##################

if ! [ $# -ge 2 ]
then
	usage
	exit 1
fi

# Check for unsupported file (neutral,signed)
FILENAME=$(basename $2 | grep -e '-signed' -e 'neutral' || true)

if [ "${FILENAME}" != "" ]
then
	echo "Error : file ${FILENAME} is either already signed or not supported neutral file"
	exit 0
fi

arg="$1"
case $arg in
	#Sign all file for a release (ipk,liveburners and burner)
	-a|--all)
	prepare_working_dir
	sign_all "$2"
	;;
	#Generate a signed burner
	-b|--burner)
	prepare_working_dir
	generate_burner "$2"
	;;
	#Display help
	-h|--help)
	usage
	;;
	# Sign a provided ipk file
	-k|--keros)
	prepare_working_dir
	sign_keros_ipk "$2" "$(dirname $2)"
	;;
	#Sign a provided liveburner file
	-l|--liveburner)
	prepare_working_dir
	sign_liveburner_ipk "$2"
	;;
	#Sign a single file (for development only): SPL, u-boot, TZ, ITB, KerOS Squash
	-f|--file)
	prepare_working_dir
	sign_file "$2"
	;;
	#Generate keros.ipk liveburner.ipk and burner from a signature package file
	-p|--package)
	prepare_working_dir
	sign_package "$2"
	;;
	*)
	usage
	exit 0
	;;
esac

#Cleanup
rm -rf ${WORKING_DIR}

exit 0
