#!/bin/bash
#

WORKING_DIR="/tmp/tool_keros"
BIN_WORKING_DIR="${WORKING_DIR}/bin"
DATA_DIR="/root/data"

# Files
#######

# IPK
IPK_ITB_FILENAME="linux.itb"
IPK_ITBS_FILENAME="${IPK_ITB_FILENAME}-signed"
IPK_CONTROL_GZ_FILENAME="control.tar.gz"
IPK_DATA_GZ_FILENAME="data.tar.gz"
IPK_DEBIANBIN_FILENAME="debian-binary"
IPK_FILES="${IPK_CONTROL_GZ_FILENAME} ${IPK_DATA_GZ_FILENAME} ${IPK_DEBIANBIN_FILENAME}"
IPK_CONTROL_FILENAME="control"
IPK_POSTINST_FILENAME="postinst"

# ITB
ITB_KERNEL_FILENAME="zImage"
ITB_DTB_FILENAME="devicetree.dtb"
ITB_INITRDGZ_FILENAME="initrd.cpio.gz"
ITB_ITS_FILENAME="fitimage.its"

# TZsw
TZSW_TZSW_FILENAME="TZsw.img"
TZSW_TZSWS_FILENAME="${TZSW_TZSW_FILENAME}-signed"

# U-Boot
UBOOT_UBOOT_FILENAME="u-boot.img"
UBOOT_UBOOTS_FILENAME="${UBOOT_UBOOT_FILENAME}-signed"
UBOOT_FITIMAGE_LOADADDR="0x87900000"

# SPL
SPL_SPL_FILENAME="SPL"
SPL_SPLS_FILENAME="${SPL_SPL_FILENAME}-signed"
SPL_SIGNINFO="SPL_signinfo"

# ROOTFS
ROOTFS_SQFS_FILENAME="keros.sqfs"
ROOTFS_SQFSS_FILENAME="${ROOTFS_SQFS_FILENAME}.sig"
INITRD_GPG_KEYS_FILE="${TOOL_DIR}/${INITRD_GPG_KEYS_FILENAME}"
INITRD_GPG_KEYS_FILENAME="${DATA_DIR}/keros_gnupg.tar.gz"

# Scripts/Bin
SCRIPTS_SIGNCSF_DIR="${TOOL_DIR}/bin"
SCRIPTS_SIGNCSF="${TOOL_DIR}/bin/signcsf.sh"
SCRIPTS_IMX_INFO="${TOOL_DIR}/bin/imxv2-info"
BIN_MKIMAGE="${TOOL_DIR}/bin/mkimage"
BIN_DUMPIMAGE="${TOOL_DIR}/bin/dumpimage"
SCRIPTS_GPG="${TOOL_DIR}/bin/signgpg.sh"

### Generic functions
##################

prepare_working_dir()
{
	echo "Preparing working dir..."

	# Clean the working dir
	rm -fR "${WORKING_DIR}"

	mkdir -p "${BIN_WORKING_DIR}"
	echo "Working dir prepared"
}


### SIGN functions
##################

#
# Sign a binary file
#
# $1: The file to sign
# $2: The signing tools arguments
#
bin_sign()
{
	local in_file="$1"
	local filename=`basename ${in_file}`
	local in_file_dir=`dirname ${in_file}`
	local ddr_addr=$(($2))

	# Calculate the size
	let image_size=$(ls -lct ${in_file} | awk '{print $5}')
	let pad_len=$((image_size - image_size % 0x1000 + 0x1000)) # align to 0x1000
	let auth_len=$((pad_len + 0x20)) # +0x20 "IVT"
	let sig_len=$((auth_len + 0x2000))
	let self_ptr=$((ddr_addr + pad_len))
	let csf_ptr=$((ddr_addr + auth_len))
	let jump_addr=$((ddr_addr + 0x1000))

	# change value to hex string
	pad_len=`printf "0x%X" ${pad_len}`
	auth_len=`printf "0x%X" ${auth_len}`
	sig_len=`printf "0x%X" ${sig_len}`
	ddr_addr=`printf "0x%X" ${ddr_addr}`
	self_ptr=`printf "0x%X" ${self_ptr}`
	csf_ptr=`printf "0x%X" ${csf_ptr}`
	jump_addr=`printf "0x%X" ${jump_addr}`

	# Select CSF template depending of type of PKI: Fast Authentication or not
	if [ -f /crts/CSF_usr_crt.pem ]
	then
			cp -f ${DATA_DIR}/csf_no_FA.tmpl ${DATA_DIR}/csf.tmpl
	else
			cp -f ${DATA_DIR}/csf_FA.tmpl ${DATA_DIR}/csf.tmpl
	fi

	# Create image.csf
	sed -e s!%ddr_addr%!${ddr_addr}!g -e s!%auth_len%!${auth_len}!g \
	    -e s!%image%!${BIN_WORKING_DIR}/${filename}!g ${DATA_DIR}/csf.tmpl > ${BIN_WORKING_DIR}/${filename}.csf

	# Create mkivt
	sed -e s!%jump_addr%!${jump_addr}!g -e s!%self_ptr%!${self_ptr}!g \
	    -e s!%csf_ptr%!${csf_ptr}!g -e s!%image%!${in_file}!g \
	    ${DATA_DIR}/mkivt.tmpl > ${BIN_WORKING_DIR}/mkivt
	chmod +x ${BIN_WORKING_DIR}/mkivt

	#######mkhabimage############
	echo "Extend ${in_file} to ${pad_len}..."
	objcopy -I binary -O binary --pad-to ${pad_len} --gap-fill=0x5A ${in_file} ${BIN_WORKING_DIR}/${filename}-pad.bin
	[ $? -ne 0 ] && exit 1

	echo "Generate IVT"
	${BIN_WORKING_DIR}/mkivt
	[ $? -ne 0 ] && exit 1

	echo "Attach IVT..."
	cat ${BIN_WORKING_DIR}/${filename}-pad.bin ivt.bin > ${BIN_WORKING_DIR}/${filename}-pad-ivt.bin
	[ $? -ne 0 ] && exit 1

	echo "Generate CSF data..."
	${SCRIPTS_SIGNCSF} ${BIN_WORKING_DIR}/${filename}.csf
	[ $? -ne 0 ] && exit 1

	echo "Merge image and CSF data..."
	cat ${BIN_WORKING_DIR}/${filename}-pad-ivt.bin ${BIN_WORKING_DIR}/${filename}.csf.bin > ${BIN_WORKING_DIR}/${filename}-signed.bin
	[ $? -ne 0 ] && exit 1

	echo "Extend final image to ${sig_len}..."
	objcopy -I binary -O binary --pad-to ${sig_len} --gap-fill=0x5A ${BIN_WORKING_DIR}/${filename}-signed.bin ${in_file}-signed
	[ $? -ne 0 ] && exit 1

	if [ ! -e ${in_file}-signed ]
	then
		echo "Error when signing file  \"${in_file}\""
		exit 1
	fi

  rm -rf ${DATA_DIR}/csf.tmpl ${BIN_WORKING_DIR}/${filename}-pad.bin ivt.bin ${filename}-pad-ivt.bin ${BIN_WORKING_DIR}/${filename}-signed.bin
}

#
# Sign an imx file
#
# $1: The file to sign
# $2: The signing tools arguments
#
imx_sign()
{
	local in_file="$1"
	local filename=`basename ${in_file}`
	local sign_args="$2"

	# Select CSF template depending of type of PKI: Fast Authentication or not
	if [ -f /crts/CSF_usr_crt.pem ]
	then
			cp -f /root/data/imx_no_FA.cfg.tpl ${BIN_WORKING_DIR}/imx.cfg
	else
			cp -f /root/data/imx_FA.cfg.tpl ${BIN_WORKING_DIR}/imx.cfg
	fi

	sed -i '/Blocks/d' ${BIN_WORKING_DIR}/imx.cfg

	# Copy binary
	cp -f ${in_file} ${BIN_WORKING_DIR}/

	# Test for mfgtool format
	local hab_status=$(echo "${sign_args}" | grep ',')
	if [ "${hab_status}" != '' ]
	then
		local hab_block1="`echo ${sign_args} | cut -d ',' -f 1`"
		local hab_block2="`echo ${sign_args} | cut -d ',' -f 2`"

		echo -e "    Blocks = ${hab_block1} \"${BIN_WORKING_DIR}/${filename}\", ${hab_block2} \"${BIN_WORKING_DIR}/${filename}\"" >> ${BIN_WORKING_DIR}/imx.cfg

		# DCD address must be cleared for signature, as mfgtool will clear it.
		dd if=${BIN_WORKING_DIR}/${filename} of=dcd_addr.bin bs=1 count=4 skip=12
		# generate a NULL address for the DCD
		dd if=/dev/zero of=zero.bin bs=1 count=4

		 # replace the DCD address with the NULL address
		dd if=zero.bin of=${BIN_WORKING_DIR}/${filename} seek=12 bs=1 conv=notrunc

		echo "Generate CSF data..."
		${SCRIPTS_SIGNCSF} ${BIN_WORKING_DIR}/imx.cfg

		# restore the DCD address with the original address
		dd if=dcd_addr.bin of=${BIN_WORKING_DIR}/${filename} seek=12 bs=1 conv=notrunc

		rm zero.bin dcd_addr.bin
	else

		echo -e "    Blocks = ${sign_args} \"/${in_file}\"" >> ${BIN_WORKING_DIR}/imx.cfg

		echo "Generate CSF data..."
		${SCRIPTS_SIGNCSF} ${BIN_WORKING_DIR}/imx.cfg
	fi

	objcopy -I binary -O binary --pad-to 0x400 --gap-fill=0x00 ${BIN_WORKING_DIR}/imx.cfg.bin ${BIN_WORKING_DIR}/imx-csf-padded.bin
	cat ${in_file} ${BIN_WORKING_DIR}/imx-csf-padded.bin > ${filename}-signed

	if [ ! -e ${in_file}-signed ]
	then
		mv ${filename}-signed ${in_file}-signed
	fi

  rm -rf ${BIN_WORKING_DIR}/imx.cfg.bin ${BIN_WORKING_DIR}/imx-csf-padded.bin ${BIN_WORKING_DIR}/imx.cfg
}

#
# Sign a file by GPG
#
# $1: The file to sign
# $2: The file to save the signature
# $3: GPG key to use
#
gpg_sign()
{
	local in_file="$1"
	local filename=`basename ${in_file}`
	local in_signature_file="$2"
	local in_gpg_key="$3"

	# Sign the file with GPG
	${SCRIPTS_GPG} ${in_file} ${in_gpg_key}
	if [ ! -e ${in_signature_file} ]
	then
		echo "Error when signing file \"${in_file}\" with GPG"
		exit 1
	fi

	echo "GPG signature file: ${in_signature_file}"

}

### ITB functions
##################

#
# Extract a keros ITB file
#
# $1: The ITB file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple itb in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
itb_extract()
{
	local in_itb_file="$1"
	local in_workspace="$2"

	# ITB directories
	local itb_dir="${WORKING_DIR}/${in_workspace}/itb"
	mkdir -p "${itb_dir}"
	local itb_kernel_file="${itb_dir}/${ITB_KERNEL_FILENAME}"
	local itb_dtb_file="${itb_dir}/${ITB_DTB_FILENAME}"
	local itb_initrdgz_file="${itb_dir}/${ITB_INITRDGZ_FILENAME}"


	echo "Extracting Linux ITB \"${in_itb_file}\" in workspace ${in_workspace}..."


	# Extracting Kernel
	${BIN_DUMPIMAGE} -i "${in_itb_file}" -T flat_dt -p 0 "${itb_kernel_file}"
	if [ $? -ne 0 ]
	then
		echo "Error when extracting kernel from ITB"
		exit 1
	fi

	# Extracting DTB
	${BIN_DUMPIMAGE} -i "${in_itb_file}" -T flat_dt -p 1 "${itb_dtb_file}"
	if [ $? -ne 0 ]
	then
		echo "Error when extracting DTB from ITB"
		exit 1
	fi

	# Extracting INITRD
	${BIN_DUMPIMAGE} -i "${in_itb_file}" -T flat_dt -p 2 "${itb_initrdgz_file}"
	if [ $? -ne 0 ]
	then
		echo "Error when extracting INITRD from ITB"
		exit 1
	fi

	echo "Linux ITB extracted"
}

#
# Build a Linux ITB from a dir with 3 elements:
# - DTB: devicetree.dtb
# - INITRD: initrd.cpio.gz
# - KERNEL: zImage
#
# $1: The workspace to use as input dir.
# $2: The final initrd file with cpio.gz extension
#
itb_build()
{
	local in_workspace="$1"
	local in_itb_file="$2"

	echo "Building Linux ITB from workspace ${in_workspace}..."
	local itb_dir="${WORKING_DIR}/${in_workspace}/itb"

	cd "${itb_dir}"
	rm -f "${in_itb_file}"
	cp -f "${DATA_DIR}/${ITB_ITS_FILENAME}" "${itb_dir}"

	${BIN_MKIMAGE} -A arm -f "${itb_dir}/${ITB_ITS_FILENAME}" "${in_itb_file}"
	if [ $? -ne 0 ]
	then
		echo "Error when creating Linux ITB"
		exit 1
	fi

	echo "Linux ITB created: ${in_itb_file}"
}

#
# Sign a Linux ITB
#
# $1: The Linux ITB file to sign
# $2: The file to save the signed ITB
#
itb_sign()
{
	local in_itb_file="$(realpath $1)"
	local in_ibts_file="$2"


	echo "Signing Linux ITB file \"${in_itb_file}\"..."
	bin_sign "${in_itb_file}" "${UBOOT_FITIMAGE_LOADADDR}"
	if [ ! -e "${in_itb_file}-signed" ]
	then
		echo "Error when signing Linux ITB"
		exit 1
	fi

	[[ "${in_itb_file}-signed" -ef "${in_ibts_file}" ]] || mv -f "${in_itb_file}-signed" "${in_ibts_file}"
	echo "Linux ITB signed: ${in_ibts_file}"
}

### TZsw functions
##################

#
# Sign a TZSW
#
# $1: The TZSW file to sign
# $2: The file to save the signed TZSW
#
tzsw_sign()
{
	local in_tzsw_file="$1"
	local in_tzsws_file="$2"

	echo "Signing TZsw file \"${in_tzsw_file}\"..."
	local TZSW_LOADADDR=$(${BIN_MKIMAGE} -l ${in_tzsw_file} | awk '/Load Address/ {printf "0x"$3"\n"}')
	bin_sign "${in_tzsw_file}" "${TZSW_LOADADDR} - 0x40"
	if [ ! -e "${in_tzsw_file}-signed" ]
	then
		echo "Error when signing TZsw"
		exit 1
	fi

	[[ "${in_tzsw_file}-signed" -ef "${in_tzsws_file}" ]] || mv -f "${in_tzsw_file}-signed" "${in_tzsws_file}"
	echo "TZsw signed: ${in_tzsws_file}"
}

### U-Boot functions
##################

#
# Sign a U-Boot
#
# $1: The U-Boot file to sign
# $2: The file to save the signed U-Boot
#
uboot_sign()
{
	local in_uboot_file="$1"
	local in_uboots_file="$2"
	echo "Signing U-Boot file \"${in_uboot_file}\"..."

	local uboot_loadaddr=$(${BIN_MKIMAGE} -l ${in_uboot_file} | awk '/Load Address/ {printf "0x"$3"\n"}')
	bin_sign "${in_uboot_file}" "${uboot_loadaddr} - 0x40"
	if [ ! -e "${in_uboot_file}-signed" ]
	then
		echo "Error when signing U-Boot"
		exit 1
	fi

	[[ "${in_uboot_file}-signed" -ef "${in_uboots_file}" ]] || mv "${in_uboot_file}-signed" "${in_uboots_file}"
	echo "U-Boot signed: ${in_uboots_file}"
}

### INITRD functions
##################

#
# Extract a cpio.gz INITRD
#
# $1: The INITRD in cpio.gz format file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
initrd_extract()
{
	in_initrd_file="$1"
	in_workspace="$2"

	echo "Extracting INITRD \"${in_initrd_file}\" in workspace ${in_workspace}..."

	# INITRD dir
	local initrd_dir="${WORKING_DIR}/${in_workspace}/initrd"
	mkdir -p "${initrd_dir}"
	cd "${initrd_dir}"
	gzip -dc "${in_initrd_file}" | cpio -idm
	if [ $? -ne 0 ]
	then
		echo "Error when extracting INITRD"
		exit 1
	fi

	echo "INITRD extracted"

}

#
# Deploy GPG keys inside an initrd
#
# $1: The tar.gz file with gpg keys to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
initrd_install_gpg_keys()
{
	local in_gpg_keys_file="$1"
	local in_workspace="$2"

	echo "Installing GPG keys \"${in_gpg_keys_file}\" in INITRD..."


	# INITRD dir
	local initrd_dir="${WORKING_DIR}/${in_workspace}/initrd"

	cd "${initrd_dir}"
	mkdir -p -m 700 root
	tar -C root -xpf "${in_gpg_keys_file}"
  chown -R root:root root
  echo "GPG keys installed"
}

#
# Build an INITRD from a dir
#
# $1: The workspace to use as input dir.
# $1: The final initrd file with cpio.gz extension
#
initrd_build()
{
	local in_workspace="$1"
	local in_initrd_file="$2"

	echo "Building INITRD from workspace ${in_workspace}..."

	local initrd_dir="${WORKING_DIR}/${in_workspace}/initrd"
	cd "${initrd_dir}"
	rm -f "${in_initrd_file}"
	find . | cpio --create --format='newc' | gzip > "${in_initrd_file}"

	echo "INITRD created: ${in_initrd_file}"

}

### IPK functions
##################

#
# Extract an IPK package
#
# $1: The IPK file to extract
# $2: The workspace to use to extract the content. It allows extraction
#     of multiple elements in the same script without conflicting.
#     e.g: keros, app_X, app_Y...
#
ipk_extract()
{
	local in_ipk="$1"
	local in_workspace="$2"

	echo "Extracting IPK file \"${in_ipk}\" in workspace ${in_workspace}..."

	# IPK directories
	local ipk_file="`basename ${in_ipk}`"
	local ipk_dir="${WORKING_DIR}/${in_workspace}/ipk"
	local ipk_data_dir="${ipk_dir}/data"
	local ipk_control_dir="${ipk_dir}/control"
	mkdir -p "${ipk_data_dir}" "${ipk_control_dir}"

	#copy the file into the workspace
	cp -f "${in_ipk}" "${ipk_dir}/"
	cd "${ipk_dir}"

	# Uncompress IPK
	ar x "${ipk_file}"

	# Uncompress data part
	tar -xzpf  "${IPK_DATA_GZ_FILENAME}" -C "${ipk_data_dir}"
	# Uncompress control part
	tar -xzpf  "${IPK_CONTROL_GZ_FILENAME}" -C "${ipk_control_dir}"

	#remove compressed files
	rm -f "${IPK_DATA_GZ_FILENAME}" "${IPK_CONTROL_GZ_FILENAME}"
	echo "IPK file \"${in_ipk}\" extracted"
}

#
# Adapt control file of IPK for a signed package
#
# $1: The workspace to use as input to adapt control file
# $2: The version of the package to force
#
ipk_adapt_control()
{
	local in_workspace="$1"
	local in_ipk_version="$2"

	local ipk_dir="${WORKING_DIR}/${in_workspace}/ipk"
	local control_file="${ipk_dir}/control/${IPK_CONTROL_FILENAME}"
	local postinst_file="${ipk_dir}/control/${IPK_POSTINST_FILENAME}"
	local arch=$(grep Architecture ${control_file} | awk '{print $2}')

	# Get SRK data from signature server
	local sign_hash=$(hexdump -n 4 -e '/4 "0x"' -e '/4 "%X""\n"' /root/cst-2.3.2/crts/SRK_1_2_3_4_fuse.bin)
	local sign_info="SRK0=${sign_hash}"
	# Adapt control file
	sed -i "s/Version: .*$/Version: ${in_ipk_version}/" "${control_file}"
	sed -i '/^Description:/,+2d' "${control_file}"
	cat >> "${control_file}" << EOF
Description: Kerlink OS
   Kerlink OS tunned for ${arch} target
   Sign information: ${sign_info}
EOF

}

#
# Build a IPK package from a list of elements:
# - debian-binary
# - data.tar.gz
# - control.tar.gz
#
# $1: The workspace to use as input to generate the IPK
# $2: The result IPK file
ipk_build()
{
	local in_workspace="$1"
	local in_ipk_file="$2"

	echo "Building IPK \"${in_ipk_file}\" from workspace ${in_workspace}..."

	local ipk_dir="${WORKING_DIR}/${in_workspace}/ipk"
	local ipk_data_dir="${ipk_dir}/data"
	local ipk_control_dir="${ipk_dir}/control"
	local ipk_data_gz_file="${ipk_dir}/${IPK_DATA_GZ_FILENAME}"
	local ipk_control_gz_file="${ipk_dir}/${IPK_CONTROL_GZ_FILENAME}"

	cd "${ipk_dir}"
	## DATA part
	#############

	# Recompress data part
	tar -czpf "${ipk_data_gz_file}" -C "${ipk_data_dir}" .

	if [ $? -ne 0 ]
	then
		echo "Error when compressing data part"
		exit 1
	fi

	## CONTROL part
	#############

	# Recompress control part
	tar -czpf "${ipk_control_gz_file}" -C "${ipk_control_dir}" .
	if [ $? -ne 0 ]
	then
		echo "Error when compressing control part"
		exit 1
	fi

	## IPK Creation
	#############
	ar r "${in_ipk_file}" ${IPK_FILES}
	if [ $? -ne 0 ]
	then
		echo "Error when creating IPK"
		exit 1
	fi

	echo "IPK built: ${in_ipk_file}"
}

### SPL functions
##################

#
# Sign a SPL
#
# $1: The SPL file to sign
# $2: The file to save the signed SPL
#
spl_sign()
{
	local in_spl_file="$1"
	local in_spls_file="$2"
	local in_splsi_file="$3"

	echo "Signing SPL file \"${in_spl_file}\"..."

	. "${in_splsi_file}"
	imx_sign "${in_spl_file}" "${SPL_HAB_BLOCKS}"
	if [ ! -e "${in_spl_file}-signed" ]
	then
		echo "Error when signing SPL"
		exit 1
	fi

	[[ "${in_spl_file}-signed" -ef "${in_spls_file}" ]] || mv "${in_spl_file}-signed" "${in_spls_file}"
	echo "SPL signed: ${in_spls_file}"
}


### Other functions
##################

#
# Load Signature information
#
# $1: The signinfo file from a neutral ipk
#
load_signinfo()
{
	local in_signinfo_file="$1"

	# Load Signature information
	if ! [ -e "${in_signinfo_file}" ]
	then
		echo "No signature information file found!"
		exit 1
	fi
	. "${in_signinfo_file}"
}
